<?php

namespace Drupal\couchbasedrupal\Cache;

use Drupal\couchbasedrupal\CouchbaseBucket as Bucket;
use Drupal\couchbasedrupal\CouchbaseManager;

abstract class CouchbaseBackendGeneric {

  /**
   * Wether to do reliable reads. "Slower" but consistent.
   *
   * @var boolean
   */
  protected $reliable = TRUE;

  /**
   * Prefix for all keys in this cache bin.
   *
   * Includes the site-specific prefix in $sitePrefix.
   *
   * @var string
   */
  protected $binPrefix;

  /**
   * Couchbase bucket
   *
   * @var Bucket
   */
  protected $bucket;

  /**
   * Query options.
   *
   * @var array
   */
  protected $options = array();

  /**
   * Name of the view for this binary in the Couchbase database
   *
   * @var string
   */
  protected $view;

  /**
   * We rely on views to do bucket wide
   * operations such as getAll() and
   * removeBin(). But views take a while
   * to update (~6 seconds) depending on many
   * factors. Use TRUE to bypass the usage
   * of views and get consistent results.
   *
   * This is used during tests. Enabling consistent
   * makes bucket wide operations much slower,
   * probably due to missing indexes.
   *
   * @var bool
   */
  protected $consistent = FALSE;

  /**
   * Current time used to validate
   * cache item expiration times.
   *
   * @var mixed
   */
  protected $requestTime;

  /**
   * Refreshes the current request time.
   *
   * Uses the global REQUEST_TIME on the first
   * call and refreshes to current time on subsequen
   * requests.
   *
   * @param int $time
   */
  public function refreshRequestTime() {
    if (empty($this->requestTime)) {
      if (defined('REQUEST_TIME')) {
        $this->requestTime = REQUEST_TIME;
        return;
      }
      if (isset($_SERVER['REQUEST_TIME_FLOAT'])) {
        $this->requestTime = round($_SERVER['REQUEST_TIME_FLOAT'], 3);
        return;
      }
    }
    $this->requestTime = round(microtime(TRUE), 3);
  }

  /**
   * Fix expiration for better precision.
   *
   * @see http://docs.couchbase.com/developer/dev-guide-3.0/doc-expiration.html
   */
  protected function FixExpiration($expire) {
    // Because time() might not be exactly the same in PHP and in Couchbase,
    // use reltive TTL() instead of absolute for small timeouts.
    $timeout = $expire - time();
    if ($timeout < 30 * 24 * 60 * 60) {
      return (int) $timeout;
    }
    return (int) $expire;
  }

  /**
   * Set if getAll() and getAllKeys() should
   * have a reliable behaviour. Slower if set to TRUE.
   *
   * @param bool $consistent
   */
  public function setConsistent($consistent) {
    $this->consistent = $consistent;
  }

  /**
   * Prepend de binary prefix.
   *
   * @param string $cid
   *   The cache item ID to prefix.
   *
   * @return string
   *   The storage key for the cache item ID.
   */
  protected function getBinKey($cid) {
    $key = $this->binPrefix . $cid;
    // If you get beyound 250 bytes the PHP driver
    // for couchbase hangs.
    if (mb_strlen($key, '8bit') >= 250) {
      // We need to preserve as much of the original
      // key as possible, so we are going to truncate
      // the exceeding characters and replace
      // them by a hash.
      $segment = substr($key, 0, 250 - 13);
      $key = $segment . CouchbaseManager::shortMd5($key);
    }
    return $key;
  }

  /**
   * Key used to store the map reduce document
   * version deployed for this document.
   *
   * @return string
   */
  protected function ViewVersionKey() {
    $view_version = 0;
    return 'couchbasedrupal-bucket-views-' . $view_version . $this->binPrefix;
  }

  /**
   * Remove any views created by this backend.
   *
   * Used when tests are finished to free
   * up the documents.
   *
   */
  public function removeViews() {
    $this->bucket->remove($this->ViewVersionKey());
    $manager = $this->bucket->manager();
    $manager->removeDesignDocument($this->binPrefix);
  }

  /**
   * We need to setup some map reduce functions
   * for each binary so that we can perform
   * fast operations on the whole binary such as
   * getAll() or invalidateAll().
   *
   * @see getAll()
   * @see getAllKeys()
   *
   * @throws \Exception
   */
  protected function EnsureViews() {

    // If the definition of the views reduce
    // function changes, update this ID to
    // make sure they are updated in the Couchbase
    // document schema.
    $key = $this->ViewVersionKey();

    // We don't want to polute the binary
    // prefix, so directly use the bucket.
    if($this->bucket->get($key)) {
      return;
    }

    $manager = $this->bucket->manager();

    #region debug code to clear all views

    $documents = $manager->getDesignDocuments();
    foreach ($documents as $name => $definition) {
      $manager->removeDesignDocument($name);
    }

    #endregion

    $dd = $manager->getDesignDocument($this->binPrefix);

    if (empty($dd)) {

      $prefixLength = strlen($this->binPrefix);

      // We use this key based map reduce to reuse this definition both
      // in the regular backend and in the raw backend - where no metadata
      // is stored in the document itself.
      $map_reduce = <<<EOF
   function (doc, meta) {
     if (meta.id.substr(0, $prefixLength) == "$this->binPrefix") {
       emit(meta.id.substr($prefixLength), doc);
     }
   }
EOF;
      $document = array(
          'views' => array(
            // To retrieve all keys in this binary.
            'full' => array(
              'map' => $map_reduce
            ),
          )
        );

      $map = json_encode($document);

      $manager->upsertDesignDocument($this->binPrefix, $map);

      // TODO: upsertDesignDocument always returns true... even if the creation fails.
      if ($manager->getDesignDocument($this->binPrefix) == FALSE) {
        throw new \Exception("Could not create view.");
      }
    }

    $this->bucket->upsert($key, TRUE);
  }


  /**
   * Returns all cached items, optionally limited by a cache ID prefix.
   *
   * @param string $prefix
   *   (optional) A cache ID prefix to limit the result to.
   *
   * @return string[]
   *   An array of Keys for this binary.
   */
  public function getAll($prefix = '') {
    if (!$this->consistent) {
      $query = \CouchbaseViewQuery::from($this->view, 'full'); //->custom(array('startkey' => "%22$prefix%22"));
      // https://forums.couchbase.com/t/upgrade-php-couchbase-2-1-0/5885/3
      $result = $this->bucket->queryView($query, NULL, TRUUE);
      return array_column($result['rows'], 'value');
    }
    else {
      $bin_key_length = strlen($this->getBinKey(''));
      $result = $this->bucket->getAllItemsByPrefix($this->getBinKey('') . $prefix);
      $cache = array();
      foreach ($result as $cid => $item) {
        if (!empty($item->error) && $item->error->code == 13) {
          continue;
        }
        $item = $this->prepareItem($item, FALSE);
        if ($item) {
          $cache[substr($cid, $bin_key_length)] = $item;
        }
      }
      unset($result);
      return $cache;
    }
  }

  /**
   * Get all the keys for the items
   * in this binary.
   *
   * @return array
   */
  public function getAllKeys() {
    if (!$this->consistent) {
      $query = \CouchbaseViewQuery::from($this->view, 'full'); //->custom(array('startkey' => "%22$prefix%22"));
      // https://forums.couchbase.com/t/upgrade-php-couchbase-2-1-0/5885/3
      $result = $this->bucket->queryView($query, NULL, TRUE);
      return array_column($result['rows'], 'value');
    }
    else {
      $result = $this->bucket->getAllKeysByPrefix($this->getBinKey(''));
      return array_values($result);
    }
  }

}
